/*
 * Copyright (C) 2003  Diez B. Roggisch [deets@web.de]
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: IterativeCollectionModel.java 114 2005-12-09 15:51:51Z christianc $
 */
package org.barracudamvc.contrib.dbroggisch.display;


import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import org.barracudamvc.core.comp.AbstractTemplateModel;
import org.barracudamvc.core.comp.IterativeModel;

/**
 * Support class to build IterativeModels
 *
 * Use this class to build IterativeModels bassed on arbitrary Collections.
 * The class will iterate over the Collection and store the current item
 * in the _current variable.
 * <p>
 * To make this class work, you will have to subclass it and provide a
 * getItem implementation.<p>
 *
 * A possible usage as anonymous inner class could look like this:
 * <code><br>
 * IterativeCollectionModel ilm = new IterativeCollectionModel() {<br>
 *   public Object getItem(Strink key) {<br>
 *     if(key.equals("MYKEY"))= {
 *       return _current;<br>
 *     }<br>
 *   }}<br>
 *<p>
 * If the Collection is of type TemplateModel, it could look like this:
 * <code><br>
 * IterativeCollectionModel ilm = new IterativeCollectionModel() {<br>
 *   public Object getItem(Strink key) {<br>
 *       return ((TemplateModel)_current).getItem(key);<br>
 *     }<br>
 *   }}<br>
 *<p>
 *
 * @author <a href="mailto:deets@web.de">Diez B. Roggisch</a>
 * @version 1.0
 */
public class IterativeCollectionModel extends AbstractTemplateModel implements IterativeModel  {

    private Collection _models;
    private String _name;
    private Iterator _it;
    /**
     * The current item of the list.
     *
     */
    protected Object _current;


    /**
     * Creates a new <code>IterativeCollectionModel</code> instance.
     *
     * @param name a <code>String</code> value
     * @param models a <code>Collection</code> value
     */
    public IterativeCollectionModel(String name, Collection models) {
        _name = name;
        _models = models;
    }

    /**
     * Creates a new <code>IterativeCollectionModel</code> instance.
     *
     * @param name a <code>String</code> value
     */
    public IterativeCollectionModel(String name) {
        this(name, new ArrayList());
    }

    /**
     * Creates a new <code>IterativeCollectionModel</code> instance.
     *
     */
    public IterativeCollectionModel() {
        this(null, new ArrayList());
    }


    /**
     * Sets the modelname
     *
     * @param name a <code>String</code> value
     */
    public void setName(String name) {
        _name = name;
    }


    /**
     * getName.
     *
     * @return a <code>String</code> value
     */
    public String getName() {
        return _name;
    }

    /**
     * Gets the value of _models
     *
     * @return the value of _models
     */
    public Collection getModels() {
        return this._models;
    }


    /**
     * Sets the value of _models
     *
     * @param arg_models Value to assign to this._models
     */
    public void setModels(Collection arg_models){
        this._models = arg_models;
    }

    /**
     * Describe <code>add</code> method here.
     *
     * @param val an <code>Object</code> value
     */
    public void add(Object val) {
        _models.add(val);
    }

    // implementation of org.barracudamvc.core.comp.IterativeModel interface


    public boolean hasNext()
    {
        return (_it != null && _it.hasNext());
    }

    public void preIterate()
    {
        _it = _models.iterator();
    }

    public void postIterate()
    {

    }

    public void loadNext()
    {
        _current = _it.next();
    }

}
