/*
 * Copyright (C) 2003  Stefan Armbruster [stefan@armbruster-it.de]
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: IterativeResultSetModel.java 114 2005-12-09 15:51:51Z christianc $
 */
package org.barracudamvc.contrib.sam.models;

import java.sql.ResultSet;
import java.sql.SQLException;

import org.apache.log4j.Logger;
import org.barracudamvc.core.comp.AbstractIterativeTemplateModel;

/** A specific ListModel implementation for java.sql.ResultSets.
 * To use this class, simply pass a ResultSet to the constructor,
 * the model iterates over all rows of the underlying SQL query.
 * A sample HTML fragment might look like
 * <code>
 * <table>
 *    <tr>
 *        <th>Username</th>
 *        <th>Password</th>
 *    </tr>
 *    <tr class="Dir::Iterate_Start.ResultSet Dir::Iterate_Next.ResultSet">
 *        <td><span class="Dir::Get_Data.ResultSet.name">abc</SPAN></td>
 *        <td><span class="Dir::Get_Data.ResultSet.password">abc</SPAN></td>
 *      </tr>
 *    <tr class="Dir::Iterate_End.ResultSet Dir::Discard">
 *      </tr>
 * </table>
 * </code>
 * @author Stefan Armbruster
 * @version $Id: IterativeResultSetModel.java 114 2005-12-09 15:51:51Z christianc $
 */
public class IterativeResultSetModel extends AbstractIterativeTemplateModel {

    protected static Logger logger = Logger.getLogger(IterativeResultSetModel.class.getName());

    /** ResultSet representing the SQL query */
    protected ResultSet resultSet;
    /** name of the model */
    protected String name;

    /** initialize the model
     * @param name Name of the model
     * @param rs ResultSet to iterate
     */
    public IterativeResultSetModel(String name, ResultSet rs) {
        this.resultSet = rs;
        setName(name);
    }

    /** get the name of the model
     * @return name of the model
     */
    public String getName() {
        return name;
    }

    /** set the name of the model
     * @param name name of the model
     */
    public void setName(String name) {
        this.name = name;
    }


    /** returns resultSet.next()
     * @return resultSet.next()
     */
    public boolean hasNext() {
        try {
            return resultSet.next();
        } catch (Exception ex) {
            logger.error(ex.getMessage(),ex);
            return false;
        }
    }

    /** get the model's value for a specific key
     * @return result of ResultSet's getString(colname)
     * @param key must match the columns of the SQL query
     */
    public Object getItem(String key) {
        try {
            return resultSet.getString(key);
        }
        catch (SQLException e) {
            logger.error(e.getMessage());
            return super.getItem(key);
        }
    }

    /** empty implementation, does nothing. Only reason for being here is to fully
     * implement IterativeModel.
     */
    public void loadNext() {
    }

}
