/*
 * Copyright (C) 2003  Christian Cryder [christianc@granitepeaks.com]
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: AbstractTemplateModel.java 114 2005-12-09 15:51:51Z christianc $
 */
package org.barracudamvc.core.comp;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.barracudamvc.core.comp.model.ModelListener;

/**
 * This class provides the abstract implementation
 * for a Template Model.
 */
public abstract class AbstractTemplateModel implements TemplateModel {

    public static final String UNDEFINED = "Undefined";

    protected ViewContext viewContext = null;
    protected List listeners = new ArrayList();

    //--------------- AbstractTemplateModel ----------------------
    /**
     * Add a listener to the template that's notified each time a change
     * to the data model occurs.
     *
     * @param ml the TemplateModelListener
     */
    public void addModelListener(ModelListener ml) {
        listeners.add(ml);
    }

    /**
     * Remove a listener
     *
     * @param ml the TemplateModelListener
     */
    public void removeModelListener(ModelListener ml) {
        listeners.remove(ml);
    }
    
    /**
     * Forwards the given notification event to all
     * <code>TemplateModelListeners</code> that registered
     * themselves as listeners for this template model.
     */
    public void fireModelChanged() {
        Iterator it = listeners.iterator();
        ModelListener ml = null;
        while (it.hasNext()) {
            ml = (ModelListener) it.next();
            ml.modelChanged(this);
        }
    }


    //--------------- AbstractTemplateModel ----------------------
    /**
     * Add a listener to the template that's notified each time a change
     * to the data model occurs.
     *
     * @param l the TemplateModelListener
     */  
//    public void addTemplateModelListener(TemplateModelListener l) {
//        listeners.add(l);
//    }

    /**
     * Remove a listener from the template.
     *
     * @param l the TemplateModelListener
     */  
//    public void removeTemplateModelListener(TemplateModelListener l) {
//        listeners.remove(l);
//    }

    /**
     * Reset the model to its initial (unprocessed) state. This 
     * is a convenience method that gets invoked prior to the 
     * entire model being rendered. You only need to override this 
     * method if you want to do something (like reset internal counters)
     * before the model is queried
     */
/*
//csc_100401.1 - this doesn't need to be in here     
    public void resetModel() {
        //nop    
    }
*/
    /**
     * process any directives. return false to indicate a the node
     * containing this directive should be skipped.
     */
    public boolean processDirective(TemplateDirective td) {
        return true;    //by default, allow all directives
    }

    //csc_030603.2 - added
    /**
     * get an item for a given template directive. Implement this
     * method if you want access to the full directive, not just the
     * String key. Note that if you implement this method, AND you
     * want to use the getItem(String key) method, then your implementation
     * of this method must be sure to call super.getItem() in order for
     * for the convenience method to get called.
     */
    public Object getItem(TemplateDirective td) {
        return getItem(td.getKeyName());
    }

    /**
     * Convenience method to get an item based on the key name (extracted
     * from the TemplateDirective).
     */
    public Object getItem(String key) {
        return getName()+"."+key+" "+UNDEFINED;
    }

    /**
     * Forwards the given notification event to all
     * <code>TemplateModelListeners</code> that registered
     * themselves as listeners for this template model.
     *
     * @param e  the event to be forwarded
     *
     * @see #addTemplateModelListener
     * @see TemplateModelEvent
     */
//    public void fireTemplateChanged(TemplateModelEvent e) {
//        Iterator it = listeners.iterator();
//        TemplateModelListener tml = null;
//        while (it.hasNext()) {
//            tml = (TemplateModelListener) it.next();
//            tml.modelChanged(e);
//        }
//    }

    /**
     * Return all the listeners for this given list model (returns 
     * a copy of the underlying List)
     *
     * @returns all the listeners for this given list model
     */
//    public List getListeners() { 
//        return new ArrayList(listeners); 
//    }

    //--------------- Contextual ---------------------------------
    /**
     * Specify the ViewContext. This method will generally be called
     * by the class that is using the model to actually render the data
     * in a view. The context will be specified prior to a render pass,
     * and the context will be reset to null after the render pass. 
     *
     * @param ivc the current ViewContext
     */
    public void setViewContext(ViewContext ivc) {
        viewContext = ivc;
    }

    /**
     * Get the current ViewContext
     *
     * @return the current ViewContext
     */
    public ViewContext getViewContext() {
        return viewContext;
    }

}