/*
 * Copyright (C) 2003  Christian Cryder [christianc@granitepeaks.com]
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: DefaultEventDispatcher.java 114 2005-12-09 15:51:51Z christianc $
 */
package org.barracudamvc.core.event;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import org.apache.log4j.Logger;
import org.barracudamvc.core.helper.servlet.BarracudaServletRequestWrapper;
import org.barracudamvc.plankton.data.CollectionsUtil;

/**
 * <p>This class is responsible for dispatching a queue of events
 * to interested listeners. If you ever want to create a custom
 * dispatching policy, this is the class to extend.
 *
 * <p>This particular implementation takes a repeating two phased
 * dispatch approach in order to provide a Model 2 flow control
 * mechanism. In order to pull this off, the incoming DispatchQueue
 * must be an instance of DefaultDispatchQueue. If it is not, an
 * EventException will be thrown.
 *
 * <p>Basically, we dispatch all the non-Response events in the
 * queue first. Listeners which handle these events loosely correspond
 * to controllers. Then, we dispatch all the Response events. Listeners
 * that handle these loosely correspond to views. If a response is required
 * and there is no response event, we look to the queue state to locate
 * a DEFAULT_RESPONSE_EVENT, which defaults to HttpResponseEvent. To
 * implement a single phase dispatch, just define the queue as not
 * requiring a response and avoid adding Response events to it.
 *
 * <p>Note that when we are dispatching, if we encounter an event that
 * implements Polymorphic, we create and dispatch parent events first.
 * If we enouncter an event that implements Exceptional and that event
 * is NOT handled, then we add it's parent event to the queue.
 *
 * <p>Note that the getEventChain method could probably be better optimized...
 */
public class DefaultEventDispatcher implements EventDispatcher {

    //public constants
    public final static String DEFAULT_RESPONSE_EVENT = "DefaultEventDispatcher.DefaultResponseEvent"; //(BaseEvent)
    public static int MAX_POLY_CHAIN_DEPTH = 15;
    public static int MAX_DISPATCH_QUEUE_DEPTH = 35;    //this prevents infinite looping
    protected static final Logger logger = Logger.getLogger(DefaultEventDispatcher.class.getName());

    //private vars
    private static int REQ_PHASE = 0;
    private static int RESP_PHASE = 1;

    /**
     * <p>Dispatch a queue of events. The incoming queue must be an instance
     * of DefaultDispatchQueue or an EventException will be thrown. This
     * means that if you're going to provide a custom event broker, you
     * may need to provide a custom dispatcher as well.
     *
     * <p>Note: this is really a repeating 2 Phase approach -- we will continue
     * to do dispatch until there are no remaining unhandled events in the queue.
     * This allows response event handlers to add non-response events back into
     * the queue (ie. for logging purposes, etc) and we are guaranteed that
     * they will get dispatched (barring an exception getting thrown)
     *
     * @param eb the event broker to be used to match events to listeners
     * @param context the event context for this whole dispatch cycle
     * @throws EventException
     */
    public void dispatchEvent(EventBroker eb, EventContext context) throws EventException {
        if (logger.isInfoEnabled()) logger.info("Attempting to dispatch context:"+context);

        //make sure the incoming queue is an instance of DefaultDispatchQueue
        DispatchQueue ieventQueue = context.getQueue();
        if (!(ieventQueue instanceof DefaultDispatchQueue)) throw new EventException("DispatchQueue is not an instance of DefaultDispatchQueue");
        DefaultDispatchQueue eventQueue = (DefaultDispatchQueue) ieventQueue;
        eventQueue.setResponseHandled(false);
        if (logger.isDebugEnabled()) {
            logger.debug("Dumping DispatchQueue prior to dispatch...");
            eventQueue.dumpEventQueueToConsole(1);
        }
        
        //recursive 2 Phase dispatch loop
        while (eventQueue.hasNextControlEvent() || eventQueue.hasNextViewEvent()) {
            //dispatch the request phase
            if (logger.isInfoEnabled()) logger.info("Dispatching REQ Phase");
            dispatch(REQ_PHASE, eb, context, eventQueue);
            
            //check to see if default response event is needed
            if (eventQueue.requiresResponse() &&
               !eventQueue.responseHandled() &&
                eventQueue.peekNextViewEvent()==null) {
                if (logger.isInfoEnabled()) logger.info("Using Default Response");
                BaseEvent defaultResponseEvent = (BaseEvent) context.getState(DEFAULT_RESPONSE_EVENT);
                eventQueue.addEvent(defaultResponseEvent);
            }
            if (logger.isDebugEnabled()) {
                logger.debug("Dumping DispatchQueue after REQ phase...");
                eventQueue.dumpEventQueueToConsole(1);
            }
            
            //dispatch the response phase
            if (logger.isInfoEnabled()) logger.info("Dispatching RESP Phase");
            dispatch(RESP_PHASE, eb, context, eventQueue);
            if (logger.isDebugEnabled()) {
                logger.debug("Dumping DispatchQueue after RESP phase...");
                eventQueue.dumpEventQueueToConsole(1);
            }
            
            //if the queue required a response and we didn't get one, throw
            //an UnhandledEventException
            if (eventQueue.requiresResponse() && !eventQueue.responseHandled()) {
                if (logger.isInfoEnabled()) logger.info("No response event handled - throwing UnhandledEventException");
                throw new UnhandledEventException("Unhandled Event - a reponse was required but no response event was handled!", context);
            }
        }
    }

    /**
     * Dispatch a queue of events
     *
     * @param dispatchPhase an internal flag to determine whether we're in REQ or RESP phase
     * @param eb the event broker to be used to match events to listeners
     * @param context the event context for this whole dispatch cycle
     * @param eventQueue the default dispatch queue
     * @throws EventException
     */
    protected void dispatch(int dispatchPhase, EventBroker eb, EventContext context, DefaultDispatchQueue eventQueue) throws EventException {
        String ext = eb.getEventExtension();
        String phaseStr = (dispatchPhase==RESP_PHASE ? "[Resp Phase] " : "[Req Phase] ");

        //dispatch all the request events in the queue:
        //iterate through all events. Once an event is handled, all
        //subsequent events in the queue will be marked as handled,
        //meaning only this listeners with NotifyAlways=true will still
        //get invoked.
        iterateQueue: while (dispatchPhase==RESP_PHASE ? eventQueue.hasNextViewEvent() : eventQueue.hasNextControlEvent()) {
            //get the next event
            BaseEvent event = (dispatchPhase==RESP_PHASE ? eventQueue.getNextViewEvent() : eventQueue.getNextControlEvent());
            event.setEventExtension(ext);
            if (logger.isDebugEnabled()) logger.debug(phaseStr+"Next event in queue:"+event);

            //get the event chain for this event
            List eventChain = getEventChain(event);
            if (logger.isDebugEnabled()) logger.debug(phaseStr+"Getting event chain: ("+eventChain.size()+" events in chain)");

            //now dispatch all the events in the chain
            Iterator it = eventChain.iterator();
            int cntr = -1;
            iterateChain: while (it.hasNext()) {
                //get the next event in the chain
                BaseEvent nextEvent = (BaseEvent) it.next();
                if (logger.isDebugEnabled()) logger.debug("Chain element "+(++cntr)+": "+nextEvent);

                //get the listeners for this event
                if (logger.isDebugEnabled()) logger.debug("Looking for event listeners...");
                List factoryList = findListeners(nextEvent, eb);
                if (factoryList==null || factoryList.size()<1) {
                    if (logger.isDebugEnabled()) logger.debug("No listeners for this event, moving on to next event");
                    continue;
                }
                if (logger.isDebugEnabled()) logger.debug("Found "+factoryList.size());
                
                //now dispatch to all listeners for this particular event
                if (logger.isDebugEnabled()) logger.debug("Dispatching to listeners...");
                try {
                    notifyListeners(nextEvent, factoryList, context);
                    if (logger.isDebugEnabled()) logger.debug("Handled:"+nextEvent.isHandled());

//csc_082302.2 - moved persistContext() logic to ApplicationGateway
/*                    
                //catch an ClientSideRedirectException     
                } catch (ClientSideRedirectException e) {
                    //save current context to session
                    if (logger.isDebugEnabled()) logger.debug("Saving context");
                    context.persistContext();

                    //rethrow the exception
                    if (logger.isDebugEnabled()) logger.debug("Rethrowing Client side redirect exception!");
                    throw e;
*/
                //catch an InterruptDispatchException     
                } catch (InterruptDispatchException e) {
                    if (logger.isDebugEnabled()) logger.debug("Dispatch interrupted!");
                    
                    //mark all events currently in the queue as handled
                    eventQueue.markEventsHandled();
                    if (logger.isDebugEnabled()) logger.debug("All events in queue marked as handled");
                    
                    //add the new event to the queue
                    BaseEvent newEvent = e.getNewEvent();
                    eventQueue.addEvent(newEvent);
                    if (logger.isDebugEnabled()) logger.debug("New event added to queue: "+newEvent);
                    
                    //break out of the iterate chain (we must do this because
                    //otherwise it's possible for the iterate to continue
                    //processing events which are technically no longer part of
                    //the event queue). This fixes a bug discovered by
                    //Filip Sarens (fsarens@janbe.jnj.com)
                    //
                    //Basically, the event chain was 6 deep; on the 5th event an
                    //InterruptDispatch was getting thrown, clearing the queue; the 6th
                    //event was still getting processed, however, which was causing
                    //2 different render events to both get processed...
                    break iterateChain;
                }
            } //iterateChain
                        
            //if we're in the response phase and the event was handled,
            //update the responseHandled flag in the queue
            if (dispatchPhase==RESP_PHASE && event.isHandled()) eventQueue.setResponseHandled(true);
            
            //if the primary event implements Exceptional and was not handled, add
            //it's parent event to the queue...
            if (!event.isHandled() && event instanceof Exceptional) {
                try {
                    BaseEvent newEvent = (BaseEvent) event.getClass().getSuperclass().newInstance();
                    newEvent.setSource(event);
                    eventQueue.addEvent(newEvent);
                    if (logger.isDebugEnabled()) logger.debug("Exceptional event not handled...adding parent event to queue: "+newEvent);
                } catch (InstantiationException ie) {
                    throw new EventException ("Error instantiating parent event:"+ie, ie);
                } catch (IllegalAccessException iae) {
                    throw new EventException ("Error instantiating parent event:"+iae, iae);
                }
            }

            //finally, check to make sure the size of the handled events has not
            //exceeded the max value. This is generally just to ensure that we
            //are not in an endless loop...
            if (eventQueue.numberOfEventsProcessed()>MAX_DISPATCH_QUEUE_DEPTH) {
                throw new UnhandledEventException("Max Dispatch Queue Depth exceeded...could indicate a recursive dispatch problem", context);
            }
            
        } //iterateQueue
    }
    
    /**
     * Get the event chain for a particular event. If this event implements
     * Polymorphic, this will be a list of events starting from the highest
     * object in the chain that implements polymorphic all the way down to the
     * actual event we pass in. If it doesn't implement polymorphic, it'll just
     * be the incoming event...
     *
     * Note: this method can probably be optimized to only instantiate when
     * there are actually listeners...
     *
     * @param event the base event for which we're trying to find an event chain
     * @return list of events in the chain
     * @throws EventException
     */
    protected List getEventChain(BaseEvent event) throws EventException {
        if (logger.isDebugEnabled()) logger.debug("Getting event chain for event:"+event);
        if (event==null) return null;
        List list = new ArrayList();

        //add the initial event
        list.add(event);
        if (event.isHandled()) return list;    //if the event has been handled, ignore Polymorphic connotations
                
        //now see if we can add any polymorphic events
        BaseEvent curEvent = event;
        Class superclazz = curEvent.getClass().getSuperclass();
        int cntr = 0;
        while ((Polymorphic.class.isAssignableFrom(superclazz)) &&
               (++cntr<MAX_POLY_CHAIN_DEPTH)) {
            //get the parent event
            BaseEvent parentEvent = null;
            try {
                parentEvent = (BaseEvent) superclazz.newInstance();
                parentEvent.setSource(curEvent);
            } catch (InstantiationException ie) {
                throw new EventException ("Error instantiating parent event:"+ie, ie);
            } catch (IllegalAccessException iae) {
                throw new EventException ("Error instantiating parent event:"+iae, iae);
            }

            if (logger.isDebugEnabled()) logger.debug("Parent event:"+parentEvent);
            
            //insert it at the beginning of the list
            list.add(0,parentEvent);
            
            //keep a reference to the cur event
            curEvent = parentEvent;
            superclazz = curEvent.getClass().getSuperclass();
        }
        if (logger.isDebugEnabled()) {
            logger.debug("Dumping Event chain...");
            CollectionsUtil.printStackTrace(list, 1, logger, null);
        }

        //now return the list
        return list;        
    }    
    
    /**
     * Find the listeners for an event
     *
     * @param event the base event for which we're trying to find listeners
     * @param eb the event broker to use when looking up listeners
     */
    protected List findListeners(BaseEvent event, EventBroker eb) {
        //see if this event is targeted for a specific group of listeners
        List factoryList = null;
        List idList = event.getListenerIDs();
        if (logger.isDebugEnabled()) logger.debug("Checking for specific listeners:"+idList);
        if (idList!=null && idList.size()>0) {
            factoryList = new ArrayList();
            Iterator it = idList.iterator();
            while (it.hasNext()) {
                Object id = it.next();
                Object factory = eb.getEventListener(id);
                if (logger.isDebugEnabled()) logger.debug("Target id:"+id+" Target listener:"+factory);
                if (factory!=null) {
                    factoryList.add(factory);
                    if (logger.isDebugEnabled()) logger.debug("Targeting listener:"+factory);
                }
            }
        }
        
        //if not, get all listeners for this particular event
        if (factoryList==null || factoryList.size()<1) {
            try {factoryList = eb.getEventListeners(event.getClass());}
            catch (InvalidClassException e) {}
            if (logger.isDebugEnabled()) logger.debug("Checking for generic listeners:"+factoryList);
        }
        
        return factoryList;
    }
    
    
    /**
     * Actually dispatch the specific event to the list of
     * listener factories
     *
     * @param event the event to be dispatched
     * @param list the list of listeners to be notified for this particular event
     * @param context the event context containing event, queue, and, sometimes, http information
     * @throws EventException
     */
    protected void notifyListeners(BaseEvent event, List list, EventContext context) throws EventException {
        //make sure we have listeners
        if (list==null || list.size()<1) {
            if (logger.isDebugEnabled()) logger.debug("List is empty, returning");
            return;
        }

        //jbh_112202.1_start
        if (context instanceof ControlEventContext) {
            HttpServletRequest req = ((ControlEventContext) context).getRequest();
            Map params = event.getParams();

//csc_010404_1_start
//            if (req!=null && params!=null &&
//                req instanceof org.barracudamvc.core.helper.servlet.HttpServletRequestWrapper) {
//                org.barracudamvc.core.helper.servlet.HttpServletRequestWrapper wreq = (org.barracudamvc.core.helper.servlet.HttpServletRequestWrapper) req;

            if (req!=null && params!=null && req instanceof BarracudaServletRequestWrapper) {
                BarracudaServletRequestWrapper wreq = (BarracudaServletRequestWrapper) req;
//csc_010404_1_end
                Iterator eventKeys = params.keySet().iterator();
                while (eventKeys.hasNext()) {
                    String key = (String) eventKeys.next();
                    Object val = params.get(key);
                    if (val instanceof String[]) {
                        String[] s = (String[]) val;
                        for (int i=0; i<s.length; i++) {
                            wreq.addParameter(key, s[i]);
                        }
                    } else {
                        wreq.addParameter(key, val.toString());
                    }
                }
            }
        }
        //jbh_112202.1_end





        //set the current event in the context (this ensures that the
        //listener always receives the correct event)
        context.putState(EventContext.BASE_EVENT, event);
        
        Iterator it = list.iterator();
        while (it.hasNext()) {
            //get the next listener
            Object o = it.next();
            if (logger.isDebugEnabled()) logger.debug("Got Next factory:"+o);
            ListenerFactory factory = (ListenerFactory) o;
            
            //see if it even needs to be notified
            if (event.isHandled() && !factory.notifyAlways()) {
                if (logger.isDebugEnabled()) logger.debug("Skipping notification because event is already handled");
                continue;
            }
            
            //get the listener
            BaseEventListener listener = factory.getInstance();
            
            //dispatch the event to the listener
            if (logger.isDebugEnabled()) logger.debug("Notifying listener:"+listener);
            listener.handleEvent(context);
        }
    }
}
