/*
 * Copyright (C) 2003  Christian Cryder [christianc@granitepeaks.com]
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: LongRunningEventGateway.java 114 2005-12-09 15:51:51Z christianc $
 */
package org.barracudamvc.core.event.events;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.barracudamvc.core.comp.AbstractTemplateModel;
import org.barracudamvc.core.comp.BAction;
import org.barracudamvc.core.comp.TemplateDirective;
import org.barracudamvc.core.event.ApplicationGateway;
import org.barracudamvc.core.event.BaseEvent;
import org.barracudamvc.core.event.BaseEventListener;
import org.barracudamvc.core.event.ClientSideRedirectException;
import org.barracudamvc.core.event.ControlEventContext;
import org.barracudamvc.core.event.DefaultBaseEventListener;
import org.barracudamvc.core.event.DefaultEventGateway;
import org.barracudamvc.core.event.DefaultListenerFactory;
import org.barracudamvc.core.event.EventException;
import org.barracudamvc.core.event.ListenerFactory;
import org.barracudamvc.core.event.LongRunning;
import org.barracudamvc.core.event.events.xmlc.LongRunningHTML;
import org.barracudamvc.core.event.helper.BTemplateViewHandler;
import org.barracudamvc.core.event.helper.EventForwardingFactory;
import org.barracudamvc.core.helper.servlet.DeferredResponseWrapper;
import org.barracudamvc.core.helper.servlet.ScriptDetector;
import org.barracudamvc.core.util.http.URLRewriter;
import org.barracudamvc.core.view.ViewCapabilities;
import org.barracudamvc.plankton.StringUtil;
import org.barracudamvc.plankton.data.ObjectRepository;


/**
 * Event handlers (both Controller and View) for the LongRunning events
 */
public class LongRunningEventGateway extends DefaultEventGateway {

    //public constants
    protected static final Logger logger = Logger.getLogger(LongRunningEventGateway.class.getName());

    //configurable through object repos assembler
    public static Class DEFAULT_TEMPLATE = LongRunningHTML.class;
    public static int DEFAULT_ETA = 180;
    public static int DEFAULT_REFRESH_RATE = 3;
    public static int LR_PIPE_SIZE = -1;                        //size of the long running pipe (-1 = default of BetterPipedReader.DEFAULT_PIPE_SIZE)
    public static int LR_PIPE_TIMEOUT = -1;                     //timeout for the long running pipe (-1 = default of BetterPipedReader.DEFAULT_TIMEOUT)
    public static int LR_DEBUG = 0; //0=normal, 1=allow resize on frames, 2=turn off the check redirect frame (useful for debugging), 3=turn of check/wait-for frames

    //Get_Data directives supported
    public static final String CHECK_LONG_RUNNING = "CheckLongRunning";                     //(BAction)
    public static final String CANCEL_LONG_RUNNING = "CancelLongRunning";                   //(BAction)
    public static final String TIME_ETA = "TimeETA";                                        //(String describing ETA in hours, minutes, and secs)
    public static final String TIME_ELAPSED = "TimeElapsed";                                //(String describing ETA in hours, minutes, and secs)
    public static final String TIME_REMAINING = "TimeRemaining";                            //(String describing ETA in hours, minutes, and secs)
    public static final String PERCENT_COMPLETE = "PercentComplete";                        //(int)
    public static final String PERCENT_REMAINING = "PercentRemaining";                      //(int)
    public static final String STATEMAP_VALUE = "StateMapValue";                            //(String - data portion of TemplateDirective used as StateMap key)
    public static final String REFRESH_RATE = "RefreshRate";                                //(int - in seconds)

    //req params
    public static final String LONG_ID = "$longid";                                         //(String) - yes, this needs to be distinct from ApplicationGateway.LR_ID

    //local vars
    public static int UID_HISTORY = 50;
    protected static List finishedUIDs = Collections.synchronizedList(new ArrayList(UID_HISTORY));
    private static long uid = System.currentTimeMillis();
    private static final String eol = System.getProperty("line.separator");

    //this defines the various event handlers
    private ListenerFactory baseLongRunningFactory =       new DefaultListenerFactory() {public BaseEventListener getInstance() {return new BaseLongRunningHandler();}        public String getListenerID() {return getID(BaseLongRunningHandler.class);}};                
    private ListenerFactory checkLongRunningFactory =      new EventForwardingFactory(new RenderLongRunningEvent());                
    private ListenerFactory cancelLongRunningFactory =     new DefaultListenerFactory() {public BaseEventListener getInstance() {return new CancelLongRunningHandler();}        public String getListenerID() {return getID(CancelLongRunningHandler.class);}};                
    private ListenerFactory renderLongRunningFactory =     new DefaultListenerFactory() {public BaseEventListener getInstance() {return new RenderLongRunningHandler();}        public String getListenerID() {return getID(RenderLongRunningHandler.class);}};                


    /**
     * Public constructor
     */
    public LongRunningEventGateway() {
        //specify who's interested in what
        specifyLocalEventInterests(baseLongRunningFactory, LongRunningEvent.class);
        specifyLocalEventInterests(checkLongRunningFactory, CheckLongRunningEvent.class);
        specifyLocalEventInterests(cancelLongRunningFactory, CancelLongRunningEvent.class);
        specifyLocalEventInterests(renderLongRunningFactory, RenderLongRunningEvent.class);
    }


    //------------------------------------------------------------
    //             Model 2 - Controller Event Handlers
    //------------------------------------------------------------
    /**
     * Gets invoked for CheckLongRunningEvent, CancelLongRunningEvent, and
     * WaitForLongRunningEvent. Basically just makes sure that everything needed is 
     * available before continuing.
     */
    class BaseLongRunningHandler extends DefaultBaseEventListener {
        public void handleControlEvent(ControlEventContext context) throws EventException, ServletException, IOException {
            //get the basic objects
            ObjectRepository lror = ObjectRepository.getObjectRepository(ApplicationGateway.LR_OR, ObjectRepository.THREADSAFE);
            ObjectRepository lor = ObjectRepository.getLocalRepository();
            HttpServletRequest req = (HttpServletRequest) lor.getState(ApplicationGateway.HTTP_SERVLET_REQUEST);
            HttpServletResponse resp = (HttpServletResponse) lor.getState(ApplicationGateway.HTTP_SERVLET_RESPONSE);

            //get the lrid
            String lrid = req.getParameter(LONG_ID);
            logger.info("invoking BaseLongRunningHandler, for "+LONG_ID+"="+lrid);

            //first and foremost, make sure the long running stuff hasn't already finished
            if (!isUIDActive(lrid)) {
                context.getQueue().setRequiresResponse(false);
                showInactiveUIDResults(lrid, resp);
                return;
            }

            //if we don't have any accompanying lrid, err
            if (lrid==null) {
                resp.sendError(HttpServletResponse.SC_GONE, "This event was missing a long running target ID. Please try running the report again.");
                return;
            } else {
                lor.putState(LONG_ID, lrid);
            }

            //if there is no DeferredResponseWrapper living in the long running OR for this lrid, err
            DeferredResponseWrapper dresp = (DeferredResponseWrapper) lror.getState(lrid);
            if (dresp==null) {
                resp.sendError(HttpServletResponse.SC_GONE, "This long running response was missing from the lror. The long running report has either completed or been canceled.");
                return;
            } else {
                lor.putState(lrid, dresp);
            }

            //sit and spin until either we have a LongRunning object available in the dresp 
            //(what happens here is the CheckLongRunningEvent handler gets dispatched at the 
            //same time as the request to actually execute the long running task - this means 
            //we may actually be handling this request before the LongRunning object has been 
            //added to the dresp yet...so if its not there, we want to sit and spin a bit until 
            //we get it; if it doesn't show up after 5 secs, something is wrong)
            BaseEvent sourceEvent = context.getEvent().getRootEvent();
            long timeout = 10000; //max 10 secs
            long millis = System.currentTimeMillis();
            while (dresp.getLongRunning()==null) {
                try {
                    if (System.currentTimeMillis()-millis>timeout) {
                        String errmsg = "This LongRunning object never showed up in the DeferredResponseWrapper. Aborting LongRunningHandler...";
                        logger.warn("Redirecting client for err: "+errmsg);
                        resp.sendError(HttpServletResponse.SC_GONE, errmsg);
                        return;
                    }
                    if (logger.isInfoEnabled()) logger.info("LongRunning object not set up yet...waiting for a bit (max:"+timeout+")");
                    Thread.sleep(300);
                    Thread.yield();
                } catch (InterruptedException e) {
                }
            }
            logger.info("long running stuff ready");
            
            //this is actually super important: we basically need to pause the dresp object (which in turn
            //will pause the write buffer to pause) - this allows this request to finish, without running into
            //a maxed-buffer situation (which results in an I/O block). We resume the dresp as soon as this
            //event is finished being handled
            if (sourceEvent instanceof CheckLongRunningEvent) dresp.pause();
        }
    }

    /**
     * Actually cancel the long running process
     */
    class CancelLongRunningHandler extends DefaultBaseEventListener {
        public void handleControlEvent(ControlEventContext context) throws EventException, ServletException, IOException {
            //get the dresp obj
            ObjectRepository lror = ObjectRepository.getObjectRepository(ApplicationGateway.LR_OR, ObjectRepository.THREADSAFE);
            ObjectRepository lor = ObjectRepository.getLocalRepository();
            HttpServletRequest req = (HttpServletRequest) lor.getState(ApplicationGateway.HTTP_SERVLET_REQUEST);
            HttpServletResponse resp = (HttpServletResponse) lor.getState(ApplicationGateway.HTTP_SERVLET_RESPONSE);
            String lrid = (String) lor.getState(LONG_ID);
            DeferredResponseWrapper dresp = (lrid!=null ? (DeferredResponseWrapper) lor.getState(lrid) : null);
            context.getQueue().setRequiresResponse(false);

            //first and foremost, make sure the long running stuff hasn't already finished
            if (!isUIDActive(lrid)) {
                showInactiveUIDResults(lrid, resp);
                return;
            }

            //get the LongRunning object so we can see where to go from here
            LongRunning lr = (dresp!=null ? (LongRunning) dresp.getLongRunning() : null);

            //stop the dresp
            if (dresp!=null) {
                if (logger.isInfoEnabled()) logger.info("Interrupting DeferredResponseWrapper...");
                dresp.interrupt();
                if (logger.isInfoEnabled()) logger.info("Interrupt complete!");
            }
            
            //clean up the long running OR
            if (lrid!=null) {
                if (logger.isInfoEnabled()) logger.info("Cleaning up lror");
                lror.removeState(lrid+"*");
            }
            
            //get the redirect url. If none is specified, simply write an error response
            resp.setContentType("text/html");
            StringBuffer sb = new StringBuffer(500);
            BaseEvent redirectEvent = (lr!=null ? lr.getRedirectEvent() : null);
            //...redirect the browser if possible
            if (redirectEvent!=null) {
                ClientSideRedirectException re = new ClientSideRedirectException(redirectEvent);
                String rdURL = URLRewriter.encodeRedirectURL(req, resp, re.getRedirectURL());
                if (logger.isInfoEnabled()) logger.info("LongRunning processes cancelled, redirecting to "+rdURL);
                sb.append("<html>"+eol);
                sb.append("  <head>"+eol);
                sb.append("    <title>Redirecting...</title>"+eol);
                sb.append("  </head>"+eol);
//                sb.append("  <body onload=\"window.top.pageFullyLoaded=true; window.top.location.replace('"+rdURL+"');\"  style=\"font-family: Georgia, Arial, Times New Roman\">"+eol);
                sb.append("  <body onload=\"window.top.location.replace('"+rdURL+"');\"  style=\"font-family: Georgia, Arial, Times New Roman\">"+eol);
                sb.append("    <p>&nbsp;<p>&nbsp;<p>&nbsp;<div style=\"text-align: center\">Cancelling request (this may take a few seconds...)</div>"+eol);
                sb.append("  </body>"+eol);
                sb.append("</html>"+eol);
            
            //...otherwise just print a basic response            
            } else {
                if (logger.isInfoEnabled()) logger.info("LongRunning processes cancelled, generating default description");
                sb.append("<html>"+eol);
                sb.append("  <head>"+eol);
                sb.append("    <title>Cancelled</title>"+eol);
                sb.append("  </head>"+eol);
                sb.append("  <body>"+eol);
                sb.append("    <h1>Cancelled.</h1>"+eol);
                sb.append("    <p>The Long Running process was cancelled."+eol);
                sb.append("  </body>"+eol);
                sb.append("</html>"+eol);
            }
            String content = sb.toString();
            Writer wr = resp.getWriter();
            wr.write(content);
            wr.close();
            resp.flushBuffer();
        }
    }


    public static void redirectBrowserToKickoffLongRunningRequest(String lrid, HttpServletRequest req, HttpServletResponse resp, ViewCapabilities vc, String ourl, String oqstr, BaseEvent event) throws IOException {
        //create a new deferred response and place it in the lror. We also want
        //to save a copy of the params so we can reconstitute those on subsquent requests
        ObjectRepository lror = ObjectRepository.getObjectRepository(ApplicationGateway.LR_OR, ObjectRepository.THREADSAFE);
        lrid = getUID();
        lror.putState(lrid, new DeferredResponseWrapper(LR_PIPE_SIZE, LR_PIPE_TIMEOUT));
        lror.putState(lrid+"_params", new HashMap(req.getParameterMap()));
        lror.putState(lrid+"_orig_url", ourl+(oqstr!=null ? "?"+oqstr : ""));
        lror.putState(lrid+"_orig_method", req.getMethod());
        lror.putState(lrid+"_event", event);
//System.out.println("storing lr objects in lror:");
//lror.printStackTrace();

        //now build the check status url
        ClientSideRedirectException rdException = new ClientSideRedirectException(new CheckLongRunningEvent());
        String rdURL = URLRewriter.encodeRedirectURL(req, resp, rdException.getRedirectURL()+"?"+LONG_ID+"="+lrid);
        rdURL = ScriptDetector.prepareRedirectURL(rdURL, vc);
        if (logger.isInfoEnabled()) logger.info("rdURL: "+rdURL);

        //now build the long running url
        String lrURL = ourl+(oqstr!=null ? "?"+oqstr : "")+(oqstr!=null ? "&" : "?")+ApplicationGateway.LR_ID+"="+lrid;
        if (logger.isInfoEnabled()) logger.info("lrURL: "+lrURL);

        //now build the wait-for url
/*        
        ClientSideRedirectException wfException = new ClientSideRedirectException(new WaitForLongRunningEvent());
        String wfURL = URLRewriter.encodeRedirectURL(req, resp, wfException.getRedirectURL()+"?"+LONG_ID+"="+lrid);
        wfURL = ScriptDetector.prepareRedirectURL(wfURL, vc);
        if (logger.isInfoEnabled()) logger.info("wfURL: "+wfURL);
*/
        //send the redirect request
        resp.setContentType("text/html");
        StringBuffer sb = new StringBuffer(500);
        sb.append("<html>"+eol);
        sb.append("  <head>"+eol);
        sb.append("    <title>Redirecting...</title>"+eol);
        sb.append("  </head>"+eol);
//        sb.append("  <script language=\"javascript\">"+eol);
//        sb.append("    var pageFullyLoaded = false;"+eol);
//This seems to be a bit quirky in practice...
//        sb.append("    function handleComplete() {"+eol);
/*
        sb.append("        if (!pageFullyLoaded) {");
        sb.append("            window.top.location.replace('CancelLongRunningEvent.event?"+ApplicationGateway.LR_ID+"="+lrid+"');");
        sb.append("            alert('In the future, please press the Cancel button to properly stop a long running process');");
        sb.append("        }");
*/
//        sb.append("    }"+eol);
//        sb.append("  </script>"+eol);
        if (LR_DEBUG==1) {              //allow resize
//            sb.append("  <frameset rows=\"200,100%\" onload=\"handleComplete();\">"+eol);
            sb.append("  <frameset rows=\"200,100%\">"+eol);
            sb.append("    <frame src=\""+lrURL+"\">"+eol);
            sb.append("    <frame src=\""+rdURL+"\">"+eol);
            sb.append("  </frameset>"+eol);
        } else if (LR_DEBUG==2) {       //don't show the check frame
//            sb.append("  <frameset rows=\"200,100%\" onload=\"handleComplete();\">"+eol);
            sb.append("  <frameset rows=\"200,100%\">"+eol);
            sb.append("    <frame src=\""+lrURL+"\">"+eol);
            sb.append("    <frame src=\"foo.html\" >"+eol);
            sb.append("  </frameset>"+eol);
        } else {                        //normal
//            sb.append("  <frameset rows=\"1,100%\" onload=\"handleComplete();\">"+eol);
            sb.append("  <frameset rows=\"1,100%\">"+eol);
            sb.append("    <frame src=\""+lrURL+"\" frameborder=\"0\" noresize>"+eol);
            sb.append("    <frame src=\""+rdURL+"\" frameborder=\"0\" noresize>"+eol);
            sb.append("  </frameset>"+eol);
        }
        sb.append("</html>"+eol);
        String content = sb.toString();
        resp.setContentLength(content.length());
        Writer wr = resp.getWriter();
        wr.write(content);
        wr.close();
        resp.flushBuffer();
    }        

    public static void redirectBrowserToGetLongRunningResults(String lrid, HttpServletResponse resp) throws IOException {
        //request the browser to redirect accordingly
        ObjectRepository lror = ObjectRepository.getObjectRepository(ApplicationGateway.LR_OR, ObjectRepository.THREADSAFE);
        String rdURL = (String) lror.getState(lrid+"_orig_url");
        lror.removeState(lrid+"_orig_url");
        if (rdURL==null) {
            resp.sendError(HttpServletResponse.SC_GONE, "orig url is missing...process probably already completed");
            return;
        }
        String sep = (rdURL.indexOf("?")>-1 ? "&" : "?");
        rdURL = rdURL+sep+ApplicationGateway.LR_ID+"="+lrid;
        if (logger.isInfoEnabled()) logger.info("redirecting to "+rdURL);
        resp.setContentType("text/html");
        resp.addHeader("Connection", "close");  //csc_040504_1 - super important: this tells the browser to close the connection, which then allows the redirect to occur immediately. If we don't do this, then the browser may sit and spin until the server finishes handling the lr request (because the browser thinks it can't reuse the connection until that happens) 
        StringBuffer sb = new StringBuffer(500);
        sb.append("<html>"+eol);
        sb.append("  <head>"+eol);
        sb.append("    <title>Redirecting...</title>"+eol);
        sb.append("    <script src=\"xlib/org/barracudamvc/core/scripts/FormControl.js\" type=\"text/javascript\"></script>"+eol);
        sb.append("  </head>");
//        sb.append("  <body onload=\"window.top.pageFullyLoaded=true; window.top.location.replace('"+rdURL+"');\">"+eol);
        sb.append("  <body onload=\"window.top.location.replace('"+rdURL+"');\">"+eol);
        sb.append("    <div style=\"text-align: center\">Your report is ready. <a href=\""+rdURL+"\" target=\"_top\">Click here</a> if your browser does not automatically redirect...</div>"+eol);
        sb.append("  </body>"+eol);
        sb.append("</html>"+eol);
        String content = sb.toString();
        resp.setContentLength(content.length());
        PrintWriter wr = resp.getWriter();
        wr.write(content);
        wr.close();
        resp.flushBuffer();
        Thread.yield();
        if (logger.isInfoEnabled()) logger.info("response written to client!");
    }
    
    public static void showInactiveUIDResults(String lrid, HttpServletResponse resp) throws IOException {
        //clean up the long running OR
        if (lrid!=null) {
            if (logger.isInfoEnabled()) logger.info("Cleaning up lror");
            ObjectRepository lror = ObjectRepository.getObjectRepository(ApplicationGateway.LR_OR, ObjectRepository.THREADSAFE);
            lror.removeState(lrid+"*");
        }
        
        //let them know its already been handled
        resp.sendError(HttpServletResponse.SC_GONE, "Long Running process "+lrid+" already completed");
        return;
    }
    
    private synchronized static String getUID() {
        return "_"+(++uid);
    }
    
    public static boolean isUIDActive(String uid) {
        return (!finishedUIDs.contains(uid));
    }
    public static void markUIDAsComplete(String uid) {
        finishedUIDs.add(0, uid);
        int size = -1;
        while ((size = finishedUIDs.size()) >= UID_HISTORY) {
            finishedUIDs.remove(size-1);
        }
    }
    

    //------------------------------------------------------------
    //                Model 2 - View Event Handlers
    //------------------------------------------------------------
    /**
     * RenderLongRunningHandler - 
     */
    class RenderLongRunningHandler extends BTemplateViewHandler {
        ObjectRepository lor = ObjectRepository.getLocalRepository();
        String lrid = (String) lor.getState(LONG_ID);
        DeferredResponseWrapper dresp = (DeferredResponseWrapper) lor.getState(lrid);
        LongRunning lr = (dresp!=null ? (LongRunning) dresp.getLongRunning() : null);

        public RenderLongRunningHandler() {
            logger.info("rendering RenderLongRunning...");
        }

        public Object getTemplateModels() {
            List models = (lr!=null ? lr.getAdditionalModels() : new ArrayList());
            models.add(new LocalTemplateModel());
            return models;
        }
        public Class getTemplateClass() {
            return (lr!=null ? lr.getTemplateClass() : DEFAULT_TEMPLATE);
        }
        public class LocalTemplateModel extends AbstractTemplateModel {

            //register the model by name
            public String getName() {return "LongRunningModel";}

            //provide items by key
            public Object getItem(TemplateDirective td) {
                String key = td.getKeyName();
                String data = td.getKeyData();
                Object item = null;
        
                //get a list of all key/total vals
                if (key.equals(CHECK_LONG_RUNNING)) {
                    BAction baction = new BAction(new CheckLongRunningEvent());
                    baction.setParam(LONG_ID, lrid);
                    item = baction;
             
                } else if (key.equals(CANCEL_LONG_RUNNING)) {
                    BAction baction = new BAction(new CancelLongRunningEvent());
                    baction.setParam(LONG_ID, lrid);
                    item = baction;

                } else if (key.equals(TIME_ETA)) {
                    item = StringUtil.getElapsedStr((long) (lr!=null ? lr.getETA()*1000 : 1000), Calendar.SECOND);

                } else if (key.equals(TIME_ELAPSED)) {
                    item = StringUtil.getElapsedStr((long) (lr!=null ? lr.getElapsed()*1000 : 1000), Calendar.SECOND);

                } else if (key.equals(TIME_REMAINING)) {
                    int remaining = (lr!=null ? lr.getETA() - lr.getElapsed() : 1000);
                    if (remaining<1) remaining = 1;
                    item = StringUtil.getElapsedStr((long) (remaining*1000), Calendar.SECOND);

                } else if (key.equals(PERCENT_COMPLETE)) {
                    item = (lr!=null ? ""+lr.getPercentComplete() : "??")+"%";
                    
                } else if (key.equals(PERCENT_REMAINING)) {
                    item = (lr!=null ? ""+(100 - lr.getPercentComplete()) :"??")+"%";
                    
                } else if (key.equals(REFRESH_RATE)) {
                    item = ""+(lr!=null ? lr.getRefreshRate() : -1);
                    
                } else if (key.equals(STATEMAP_VALUE)) {
                    Object val = (lr!=null ? lr.getStateMap().getState(data) : null);
                    item = (val!=null ? val : "[Undefined Statemap Key: "+data+"]");
                    
                } else {
                    item = super.getItem(td);
                }
                
                //return item
                logger.info("getItem() - key: "+key+" item:"+item);
                return item;
            }
        }
        public void cleanup() {
            if (dresp!=null) dresp.resume();    //super important - allows the dresp to start writing data again
        }
    }
}












































