/*
 * Copyright (C) 2003  Christian Cryder [christianc@granitepeaks.com]
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: DefaultFormMap.java 114 2005-12-09 15:51:51Z christianc $
 */
package org.barracudamvc.core.forms;

import java.util.*;

import javax.servlet.ServletRequest;

import org.apache.log4j.Logger;
import org.barracudamvc.plankton.data.DefaultStateMap;
import org.barracudamvc.plankton.data.ServletRequestParameterStateMap;
import org.barracudamvc.plankton.data.StateMap;

/**
 * <p>This class provides the default implementation of a FormMap.
 *
 * <p>A FormMap is used to provide a virtual representation of a
 * form. It can contain any number of unique FormElements, and
 * it can also be associated with FormValidators. The primary
 * function of a form map is to:
 *
 * <ul>
 *   <li>define the map (with its elements and validators)</li>
 *   <li>actually populate the map (from either a ServletRequest or a StateMap)</li>
 *   <li>validate the map (by invoking all the validators associated with the 
 *       form and all its elements)</li>
 *   <li>provide convenience methods to access the underlying values of the form 
 *       elements contained in this map</li>
 * </ul>
 *
 * The FormMap uses a pluggable FormMapper to actually control how elements 
 * are mapped.
 *
 * @author  Christian Cryder [christianc@granitepeaks.com]
 * @author  Diez B. Roggisch <diez.roggisch@artnology.com>
 * @author  Jacob Kjome <hoju@visi.com>
 * @version %I%, %G%
 * @since   1.0
 */
public class DefaultFormMap implements FormMap {

    //public constants
    protected static final Logger localLogger = Logger.getLogger(DefaultFormMap.class.getName());

//    private static final String PARSE_EXCEPTION = "ParseException";

    //non-public vars
    protected Map elements = new TreeMap();
    protected List validators = new ArrayList(5);
    protected StateMap statemap = new DefaultStateMap();
    protected static final Locale defaultLoc = Locale.getDefault();
    protected Locale zloc = null;

    /**
     * @clientCardinality 1 
     */
    protected FormMapper formMapper = null;

    /**
     * Default constructor (uses DefaultFormMapper)
     */
    public DefaultFormMap() {
        //nop
    }

    /**
     * Prefix constructor (uses PrefixFormMapper)
     *
     * @param prefix the prefix to be used by the PrefixFormMapper
     */
    public DefaultFormMap(String iprefix) {
        this.setFormMapper(new PrefixFormMapper(iprefix));
    }


    
    //--------------- FormMap ------------------------------------
    /**
     * This defines an element to be mapped by this form, using
     * the key from the FormElement. You would invoke this method
     * for each element in the form.
     *
     * @param element a FormElement to be mapped by this form
     * @return a reference to the newly defined FormElement
     */
    public FormElement defineElement(FormElement element) {
        return this.defineElement(element.getKey(), element);
    }

    /**
     * This defines an element to be mapped by this form, using 
     * a manually specified key. Generally, you won't need to use 
     * this method unless you are writing some kind of framwork that
     * maps multiple elements into a for (ie. with different variations
     * on a name), all via the same fundamental key.
     *
     * @param key a manually specified key
     * @param element a FormElement to be mapped by this form
     * @return a reference to the newly defined FormElement
     */
    public FormElement defineElement(String key, FormElement element) {
        if (key==null) key = element.getKey();
        if (key==null) return element; //eliminate the obvious (TreeMap doesn't like calls to get(null))
        if (localLogger.isDebugEnabled()) localLogger.debug("Defining FormElement: Key=["+key+"] Element=["+element+"]");
        elements.put(key, element);
        element.setParentForm(this);
        return element;
    }


    /**
     * This defines a validator for the entire form. This validator
     * will be invoked prior to validating specific form element
     * validators. Calling this method multiple times will result
     * in multiple form validators being added to the form (they
     * will be invoked in the order they were added)
     *
     * @param validator a form validator to be applied to the entire form
     * @return a reference to the newly defined FormValidator
     */
    public FormValidator defineValidator(FormValidator validator) {
        if (!validators.contains(validator)) {
            if (localLogger.isDebugEnabled()) localLogger.debug("Defining validator: Validator=["+validator+"]");
            validators.add(validator);
        }
        return validator;
    }

    /**
     * This is where we actually take an incoming form (in
     * the form of a ServletRequest) and map it using the
     * definitions supplied by all the FormElements. If there
     * are multiple parameters for a given key, the values
     * will be mapped into List structures
     *
     * @param req the ServletRequest to map paramters from based on
     *        all defined FormElements
     * @return a reference to the FormMap (we do this so you can inline
     *        map/validate requests if you desire: form.map(req).validate())
     * @throws MappingException if for some reason the value cannot
     *        be mapped successfully
     */
    public FormMap map(ServletRequest req) {
        if (localLogger.isInfoEnabled()) localLogger.info("Mapping ServletRequest to FormMap");
        return getFormMapper().mapForm(this, new ServletRequestParameterStateMap(req).getStateStore());        
    }

    /**
     * This is where we actually take an incoming form (in
     * the form of a StateMap) and map it using the definitions
     * supplied by all the FormElements.
     *
     * @param map the StateMap to map properties from based on
     *        all defined FormElements
     * @return a reference to the FormMap (we do this so you can inline
     *        map/validate requests if you desire: form.map(req).validate())
     * @throws MappingException if for some reason the value cannot
     *        be mapped successfully
     */
    public FormMap map(StateMap map) {
        if (localLogger.isInfoEnabled()) localLogger.info("Mapping StateMap to FormMap");
        return getFormMapper().mapForm(this, map.getStateStore());        
    }

    /**
     * Map an individual element by specifiying the key and origVal. Normally, 
     * you'll just map the whole form all at once, but if for some reason you 
     * desire to manually map an element at a time, this method gives you the 
     * ability to do that.
     *
     * @param key the form map key
     * @param origVal the orig val to be mapped
     * @return a reference to the FormElement which got mapped (null if it mapper doesn't handle this key)
     * @throws MappingException if for some reason the value cannot be mapped successfully
     */
    public FormElement mapElement(String key, Object origVal) {
        if (key==null) return null; //eliminate the obvious (TreeMap doesn't like calls to get(null))
        if (localLogger.isInfoEnabled()) localLogger.info("Mapping key:"+key+", val:"+origVal+" to FormMap");
        return getFormMapper().mapElement(this, key, origVal);
    }

    /**
     * Validate the entire form (both form level and elements).
     * We start by invoking form validators which apply to individual
     * form elements, then we invoke any which apply to the entire form
     *
     * @param deferExceptions do we want to deferValidation exceptions
     *        and attempt to validate all elements so that we can process
     *        all the exceptions at once
     * @return a reference to the FormMap (we do this so you can inline
     *        map/validate requests if you desire: form.map(req).validate())
     * @throws ValidationException if the form (or any element within it)
     *        is invalid
     */
    public FormMap validate(boolean deferExceptions) throws ValidationException {
        if (localLogger.isInfoEnabled()) localLogger.info("Validating FormMap (form & elements)");
        ValidationException ve = null;

        try {
            //validate each individual element
            try {
                validateElements(deferExceptions);
            } catch (DeferredValidationException dve) {
                if (ve==null) ve = new DeferredValidationException(dve.getSource(), "Validation err:"+dve);
                ve.addSubException(dve);
            }

            //validate the entire form
            try {
                validateForm(deferExceptions);
            } catch (DeferredValidationException dve) {
                if (ve==null) ve = new DeferredValidationException(dve.getSource(), "Validation err:"+dve);
                ve.addSubException(dve);
            }
        } catch (ValidationException e) {
            if (localLogger.isDebugEnabled()) localLogger.debug("Validation err! (immediate): ", e);
            ErrorManager.handleError(e);    //csc_110304_1
            throw e;
        }

        //now, if we have generated a ValidationExceptions,
        //rethrow it
        if (ve!=null) {
            ErrorManager.handleError(ve);   //csc_110304_1
            throw ve;
        }
        return this;
    }

    /**
     * Validate just the elements (not the form)
     *
     * @param deferExceptions do we want to deferValidation exceptions
     *        and attempt to validate all elements so that we can process
     *        all the exceptions at once
     * @return a reference to the FormMap (we do this so you can inline
     *        map/validate requests if you desire: form.map(req).validate())
     * @throws ValidationException if the form (or any element within it)
     *        is invalid
     */
    public FormMap validateElements(boolean deferExceptions) throws ValidationException {
        if (localLogger.isInfoEnabled()) localLogger.info("Validating FormMap (elements)");
        ValidationException ve = null;

        try {
            //see if we are going to require the elements to have a certain infix - basically, in the case of a UIDFormMapper,
            //we only want to consider elemetns that were actually mapped (and thus have the :: in their key)
            String infix = null;
            if (formMapper!=null && formMapper instanceof UIDFormMapper) infix = UIDFormMapper.UID_TOKEN;

            //validate each individual element
            if (localLogger.isDebugEnabled()) localLogger.debug("Validating individual elements...");
            Iterator it = elements.values().iterator();
            FormElement element = null;
            while (it.hasNext()) {
                try {
                    element = (FormElement) it.next();
                    if (infix!=null && element.getKey().indexOf(infix)<0) {
                        if (localLogger.isDebugEnabled()) localLogger.debug("Skipping FormElement: "+element+" because it lacks infix "+infix);
                        continue;
                    }    
                    if (localLogger.isDebugEnabled()) localLogger.debug("Next FormElement: "+element);
                    FormValidator fv = element.getValidator();
                    if (fv!=null) fv.validate(element, this, deferExceptions);
                    if (localLogger.isDebugEnabled()) localLogger.debug("Element Valid!");
                } catch (DeferredValidationException dve) {
                    if (localLogger.isDebugEnabled()) localLogger.debug("Element Invalid! (deferred): "+dve.getMessage());
                    if (ve==null) ve = new DeferredValidationException(element, "Validation err:"+dve);
                    ve.addSubException(dve);
                }
            }
        } catch (ValidationException e) {
            if (localLogger.isDebugEnabled()) localLogger.debug("Validation err! (immediate): ", e);
            throw e;
        }

        //now, if we have generated a ValidationExceptions,
        //rethrow it
        if (ve!=null) throw ve;
        return this;
    }

    /**
     * Validate just the form (not the individual elements)
     *
     * @param deferExceptions do we want to deferValidation exceptions
     *        and attempt to validate all elements so that we can process
     *        all the exceptions at once
     * @return a reference to the FormMap (we do this so you can inline
     *        map/validate requests if you desire: form.map(req).validate())
     * @throws ValidationException if the form (or any element within it)
     *        is invalid
     */
    public FormMap validateForm(boolean deferExceptions) throws ValidationException {
        if (localLogger.isInfoEnabled()) localLogger.info("Validating FormMap (form)");
        ValidationException ve = null;

        try {
            //validate the entire form
            if (localLogger.isDebugEnabled()) localLogger.debug("Validating entire form...");
            try {
                Iterator it = validators.iterator();
                while (it.hasNext()) {
                    Object obj = it.next();
                    if (localLogger.isDebugEnabled()) localLogger.debug("Next obj:"+obj);
                    FormValidator fv = (FormValidator) obj;
                    if (localLogger.isDebugEnabled()) localLogger.debug("Next FormValidator: "+fv);
                    fv.validate(null, this, deferExceptions);
                    if (localLogger.isDebugEnabled()) localLogger.debug("Form Valid!");
                }
            } catch (DeferredValidationException dve) {
                if (localLogger.isDebugEnabled()) localLogger.debug("Form Invalid! (deferred): ", dve);
                if (ve==null) ve = new DeferredValidationException(this, "Validation err:"+dve);
                ve.addSubException(dve);
            }
        } catch (ValidationException e) {
            if (localLogger.isDebugEnabled()) localLogger.debug("Validation err! (immediate): ", e);
            throw e;
        }

        //now, if we have generated a ValidationExceptions,
        //rethrow it
        if (ve!=null) throw ve;
        return this;
    }

    /**
     * Return true if an element exists and its value is not null
     *
     * @param key the key which uniquely identifies this FormElement
     * @return true if an element exists (not null)
     */
    public boolean exists(String key) {
        if (key==null) return false; //eliminate the obvious (TreeMap doesn't like calls to get(null))
        FormElement fel = (FormElement) elements.get(key);
        if (fel==null) return false;
        return (fel.getVal()!=null);
    }

    /**
     * Get an element by key
     *
     * @param key the key which uniquely identifies this FormElement
     * @return the FormElement for this key (may be null)
     */
    public FormElement getElement(String key) {
        if (key==null) return null; //eliminate the obvious (TreeMap doesn't like calls to get(null))
        return (FormElement) elements.get(key);
    }

    /**
     * return a List with containing all the defined elements in this FormMap
     *
     * @return a copy of the defined elements in this FormMap
     */
    public Map getElements() {
        return new TreeMap(elements);
    }

    /**
     * Manually set the value of an element. If an element does not
     * exist for this key one will be created.
     *
     * @param key the key
     * @param val the value for the key
     */
    public void setVal(String key, Object val) {
        if (key==null) return; //eliminate the obvious (TreeMap doesn't like calls to get(null))
        FormElement el = (FormElement) elements.get(key);
        if (el==null) {
            el = new DefaultFormElement(key);
            this.defineElement(el);
        }
        el.setVal(val);
    }

    /**
     * Get the value for a given key. This is basically a convenience
     * method. You could manually grab the FormElement using getElement
     * and then retrieve the value that way as well. Note that if the
     * particular FormElement supports multiple values, then this call
     * will only return the first value in the array; to get all the values,
     * use the getVals() function.
     *
     * @param key the key which uniquely identifies this FormElement
     * @return the value for this key (may be null)
     */
    public Object getVal(String key) {
        return getVal(key, null);
    }

    public Object getVal(String key, Object dflt) {
        if (key==null) return null; //eliminate the obvious (TreeMap doesn't like calls to get(null))
        FormElement el = (FormElement) elements.get(key);
        if (el==null) return null;
        Object val = el.getVal();
        if (val==null) val = dflt;
        if (val==null) val = el.getDefaultVal();
        return val;
    }

    /**
     * Get an array of values for a given key. This is basically a convenience
     * method. You could manually grab the FormElement using getElement
     * and then retrieve the value that way as well. You should only use this
     * method if the particular FormElement has allowMultiples = true
     *
     * @param key the key which uniquely identifies this FormElement
     * @return the value for this key (may be null)
     */
    public Object[] getVals(String key) {
        return getVals(key, null);
    }
    
    public Object[] getVals(String key, Object[] dflt) {
        if (key==null) return null; //eliminate the obvious (TreeMap doesn't like calls to get(null))
        FormElement el = (FormElement) elements.get(key);
        if (el==null) return null;
        Object[] vals = el.getVals();
        return (vals!=null ? vals : dflt);
    }

    /**
     * Set the working locale for mapping the elements (null indicates the default 
     * locale will be used)
     *
     * @param loc the working locale
     */
    public void setLocale(Locale iloc) {
        zloc = iloc;
    }
    
    /**
     * Get the working locale for this form
     *
     * @returns the working locale for this form
     */
    public Locale getLocale() {
        return (zloc!=null ? zloc : defaultLoc);    
    }

    /**
     * Set the FormMapper for this map
     */
    public void setFormMapper(FormMapper imapper) {
        formMapper = imapper;
    }
    
    /**
     * Get the FormMapper for this map
     */
    public FormMapper getFormMapper() {
        if (formMapper==null) formMapper = new DefaultFormMapper();
        return formMapper;
    }




    //-------------------- StateMap ------------------------------
    /**
     * set a property in this StateMap
     *
     * @param key the state key object
     * @param val the state value object
     */
    public void putState(Object key, Object val) {
        statemap.putState(key,val);
    }

    /**
     * get a property in this StateMap
     *
     * @param key the state key object
     * @return the value for the given key
     */
    public Object getState(Object key) {
        return statemap.getState(key);
    }

    /**
     * remove a property in this StateMap
     *
     * @param key the key object
     * @return the object which was removed
     */
    public Object removeState(Object key) {
        return statemap.removeState(key);
    }

    /**
     * get a keyset for this StateMap (whether or 
     * not the set is backed by the data store depends on 
     * the implementation)
     *
     * @return a Set of keys for this StateMap
     */
    public Set getStateKeys() {
        return statemap.getStateKeys();
    }

    /**
     * get a copy of the underlying Map
     *
     * @return a copy of the underlying state Map
     */
    public Map getStateStore() {
        return statemap.getStateStore();
    }

    //csc_052803_2 - added
    /**
     * clear all state information
     */
    public void clearState() {
        statemap.clearState();
    }
}
