/*
 * Copyright (C) 2001  Christian Cryder [christianc@granitepeaks.com]
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: DefaultDOMWriter.java 114 2005-12-09 15:51:51Z christianc $
 */
package org.barracudamvc.core.util.dom;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;

import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.enhydra.xml.io.DOMFormatter;
import org.enhydra.xml.io.OutputOptions;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.html.HTMLDocument;
import org.w3c.tidy.Tidy;


/**
 * This class provides the default implementation for writing a DOM.
 * XMLC is used for formatting the dom.  Besides output options features
 * supplied by XMLC, supported features include pretty printing and preventing
 * page caching (which prevents any redisplay of the page even via the back
 * button) or setting a max-age (which generally allows for redisplay of a
 * page via the back button, but not when revisiting the URL).  So, the prevent
 * caching and max age features are mutually exclusive. Default behavior is for
 * default output options (obtained automatically from the current document),
 * no pretty printing, and setting a max-age header of 0 (zero).  Defaults may
 * be modified via class static variables or overridden explicitly via
 * constructors and/or mutators.
 */
public class DefaultDOMWriter implements DOMWriter {

    protected static final Logger logger = Logger.getLogger(DefaultDOMWriter.class.getName());

    //csc_111103_1 - added
    /**
     * default OutputOptions public id (ie. "-//W3C//DTD HTML 4.01 Transitional//EN"); if
     * set, the default output options will use this value
     */
    public static String DEFAULT_OO_PUBLIC_ID = null;

    //csc_111103_1 - added
    /**
     * default OutputOptions system id (ie. "http://www.w3.org/TR/html401/loose.dtd"); if
     * set, the default output options will use this value
     */
    public static String DEFAULT_OO_SYSTEM_ID = null;

    /**
     * default value for pretty printing, false unless modified at runtime
     */
    public static boolean DEFAULT_PRINT_PRETTY = false;

    /**
     * default value for preventing caching, false unless modified at runtime
     */
    public static boolean DEFAULT_PREVENT_CACHING = false;

    /**
     * default value for max age of rendered page, 0 (zero) unless modified at runtime
     */
    public static int DEFAULT_MAX_AGE = 0;

    protected DOMFormatter dfm = null;
    protected OutputOptions oo = null;
    protected String contentType = null;                    //csc_012804_1
    protected String contentDisposition = null;             //csc_012804_1
    protected boolean printPretty = false;
    protected boolean preventCaching = false;
    protected boolean leaveWriterOpen = false;
    protected int maxAge = 0;   //csc_061202.1 - added

    /**
     * Default constructor
     */
    public DefaultDOMWriter() {
        this(null, DEFAULT_PRINT_PRETTY, DEFAULT_PREVENT_CACHING, DEFAULT_MAX_AGE);
    }

    /**
     * Public constructor. Allows you to specify OutputOptions.
     *
     * @param oo OutputOptions to specify how the DOM should be formatted
     */
    public DefaultDOMWriter(OutputOptions oo) {
        this(oo, DEFAULT_PRINT_PRETTY, DEFAULT_PREVENT_CACHING, DEFAULT_MAX_AGE);
    }

    /**
     * Public constructor. Allows you to specify pretty printing.
     *
     * @param printPretty true if pretty print
     */
    public DefaultDOMWriter(boolean printPretty) {
        this(null, printPretty, DEFAULT_PREVENT_CACHING, DEFAULT_MAX_AGE);
    }

    /**
     * Public constructor. Allows you to specify output options,
     * pretty printing and whether or not the page should be cached or the
     * max-age header for the page (how many seconds until it expires).
     * Note that the latter two options are mutually exclusive.
     *
     * @param ioo OutputOptions to specify how the DOM should be formatted, may be null
     * @param iprintPretty true if pretty print
     * @param ipreventCaching true if we want to prevent client side caching of the page
     * @param imaxAge how many seconds until the page should expire
     */
    public DefaultDOMWriter(OutputOptions ioo, boolean iprintPretty, boolean ipreventCaching, int imaxAge) {
        logger.info("Instantiating DOMWriter: "+this);
        setOutputOptions(ioo);
        setPrettyPrint(iprintPretty);
        setPreventCaching(ipreventCaching);
        setMaxAge(imaxAge);
    }

//csc_012804_1_start
    /**
     * Set the content type (defaults to "text/html" or "text/xml" depending on the 
     * document type
     */
    public void setContentType(String icontentType) {
        contentType = icontentType;
    }

    /**
     * Get the content type
     */
    public String getContentType() {
        return contentType;
    }

    /**
     * Set the content disposition (ie. "inline; filename=foo.txt", defaults to null)
     */
    public void setContentDisposition(String icontentDisposition) {
        contentDisposition = icontentDisposition;
    }

    /**
     * Get the content disposition
     */
    public String getContentDisposition() {
        return contentDisposition;
    }

    /**
     * Set true if we want to leave the writer open (ie. for multiple writes)
     */
    public void setLeaveWriterOpen(boolean val) {
        leaveWriterOpen = val;
    }

    /**
     * return true if the writer is configured to leave the output stream open
     */
    public boolean getLeaveWriterOpen() {
        return leaveWriterOpen;
    }

    /**
     * Prepare the response object
     *
     * @param node the DOM node to be written out
     * @param resp the HttpServletResponse object
     */
    public void prepareResponse(Node node, HttpServletResponse resp) throws IOException {
        //saw_031604_1 - set a default content type if none is specified
        if (getContentType()==null) setContentType((node instanceof HTMLDocument) ? "text/html" : "text/xml");
        
        //set the content type and disposition
        if (logger.isDebugEnabled()) logger.debug("Setting content type:"+contentType+" content disposition:"+contentDisposition);
        if (contentType!=null) resp.setContentType(contentType);
        if (contentDisposition!=null) resp.setHeader("Content-Disposition", contentDisposition);

        //if we need to prevent caching
        if (preventCaching) {
            //add the appropriate headers to the response
            if (logger.isDebugEnabled()) logger.debug("updating resp hdr to prevent caching");
            resp.setHeader("Pragma","no-cache");
            resp.setHeader("Cache-Control","no-cache");
            resp.setDateHeader("Expires", System.currentTimeMillis());

        //otherwise explicitly give it a max-age (this will generally allow browsers like
        //IE to page back in history without reloading, but if the user actually revisits the
        //URL, then it will still be reloaded)
        } else {
            resp.setHeader("Cache-Control","max-age="+maxAge);
            resp.setDateHeader("Last-Modified", System.currentTimeMillis());
        }
    }
//csc_012804_1_end

    /**
     * Write a DOM to a ServletResponse object. This method will
     * automatically set the content type for you.
     *
     * @param node the DOM node to be written out
     * @param resp the HttpServletResponse object
     */
    public void write(Node node, HttpServletResponse resp) throws IOException {
//csc_012804_1_start
/*
        //set the response type (TODO:need to support WML as well)
        String responseType = "text/xml";
        if (node instanceof HTMLDocument) responseType = "text/html";
        if (logger.isDebugEnabled()) logger.debug("Setting the response type:"+responseType);
        resp.setContentType(responseType);

        //if we need to prevent caching
        if (preventCaching) {
            //add the appropriate headers to the response
            if (logger.isDebugEnabled()) logger.debug("updating resp hdr to prevent caching");
            resp.setHeader("Pragma","no-cache");
            resp.setHeader("Cache-Control","no-cache");
            resp.setDateHeader("Expires", System.currentTimeMillis());

        //csc_061202.1 - added
        //otherwise explicitly give it a max-age (this will generally allow browsers like
        //IE to page back in history without reloading, but if the user actually revisits the
        //URL, then it will still be reloaded)
        } else {
            resp.setHeader("Cache-Control","max-age="+maxAge);
            resp.setDateHeader("Last-Modified", System.currentTimeMillis());
        }
*/        
//saw_031604_1        if (getContentType()==null) setContentType((node instanceof HTMLDocument) ? "text/html" : "text/xml");
        prepareResponse(node, resp);
//csc_012804_1_end

        //write the dom
        
      //saw_031604_1 begin
        // We want to use resp.getOutputStream() instead of resp.getWriter() because the PrintWriter
        // returned by getWriter() apparently has no buffer whereas the ServletOutputStream does.
        // This means that using the PrintWriter would cause the message chunks in the response to
        // be extremely small, with each call to print() causing a new chunk to be sent.
        
//        write(node, resp.getWriter());
        write(node, new OutputStreamWriter(resp.getOutputStream()));
      //saw_031604_1 end
    }

    /**
     * Write a DOM to an OutputStream.
     *
     * @param node the DOM node to be written out
     * @param out the OutputStream to be written to
     */
    public void write(Node node, OutputStream out) throws IOException {
      //saw_031604_1 begin
        // use OutputStreamWriter instead of PrintWriter... there is not need to introduce
        // buffering here. In fact, using PrintWriter as it was used here could cause some
        // output not to get printed because the gc may reclaim the PrintWriter before output is
        // flushed.

//        write(node, new PrintWriter(out));
        write(node, new OutputStreamWriter(out));
      //saw_031604_1 end
    }

    /**
     * Write a DOM to a Writer.
     *
     * @param node the DOM node to be written out
     * @param writer the writer to be written to
     */
    public void write(Node node, Writer writer) throws IOException {
        //build the default DOMFormatter if necessary
        if (dfm==null) dfm = new DOMFormatter();

        OutputOptions localoo = null;
        //set the default output options if necessary
        if (oo != null) localoo = oo;
        if (localoo == null) localoo = getDefaultOutputOptions(node.getOwnerDocument());

        //configure the formatter
        dfm.setOutputOptions(localoo);

        //now print...
        if (!printPretty) {
            //write the doc directly to the writer
            if (logger.isDebugEnabled()) logger.debug("Writing DOM directly to writer");
            dfm.write(node, writer);
            if (logger.isDebugEnabled()) logger.debug("Finished writing DOM!");
            if (!leaveWriterOpen) {
                logger.debug("Closing writer");
                writer.close();
            }
        } else {
            //NOTE: pretty printing DOES NOT WORK with BlockIteration (I haven't exactly figured out why,
            //but if you configure the DOM writer to print pretty, block iteration gets hosed - you will
            //end up with invalid markup) - csc_050604
        
            //write the doc to a byte stream
            byte[] bytes = dfm.toBytes(node);

            //filter the byte stream to tidy
            if (logger.isDebugEnabled()) logger.debug("Formatting for pretty print");
            Tidy tidy = new Tidy();
            tidy.setIndentContent(true);
            tidy.setSpaces(2);
            tidy.setQuiet(true);
            ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            tidy.parse(bais, baos);

            //now write the file
            if (logger.isDebugEnabled()) logger.debug("Writing DOM to writer");
            writer.write(baos.toString());
            if (logger.isDebugEnabled()) logger.debug("Finished writing DOM!");
            if (!leaveWriterOpen) writer.close();
        }
    }

    public void setPrettyPrint(boolean val) {
        this.printPretty = val;
    }

    public void setPreventCaching(boolean val) {
        this.preventCaching = val;
    }

    //csc_061202.1 - added
    public void setMaxAge(int imax) {
        this.maxAge = imax;
    }

    //jrk_20030317.1
    public void setOutputOptions(OutputOptions ioo) {
        this.oo = ioo;
    }
    
    /**
     * @since   saw_020204_1
     */
    public OutputOptions getOutputOptions() {
        return oo;
    }

    public static OutputOptions getDefaultOutputOptions(Document doc) {
//csc_111103_1_start
//        return DOMFormatter.getDefaultOutputOptions(doc);
        OutputOptions doo = DOMFormatter.getDefaultOutputOptions(doc);
        if (DEFAULT_OO_PUBLIC_ID!=null || DEFAULT_OO_SYSTEM_ID!=null) doo.setOmitDocType(false);
        if (DEFAULT_OO_PUBLIC_ID!=null) doo.setPublicId(DEFAULT_OO_PUBLIC_ID);
        if (DEFAULT_OO_SYSTEM_ID!=null) doo.setSystemId(DEFAULT_OO_SYSTEM_ID);
        return doo;
//csc_111103_1_end
    }

/*
    public static void main(String args[]) {
        //setup
        DOMWriter dw = new DefaultDOMWriter();
        org.enhydra.xml.xmlc.XMLCFactory xmlcFactory = new org.enhydra.xml.xmlc.XMLCStdFactory(dw.getClass().getClassLoader(), new org.enhydra.xml.xmlc.StreamXMLCLogger());
        org.barracudamvc.tutorials.xmlc.Visibility1HTML page1 = null;
        org.barracudamvc.tutorials.xmlc.Visibility2HTML page2 = null;
        StringWriter sw = null;

        //load & print DOM page
System.out.println ("");
System.out.println ("Here 0");
        page1 = (org.barracudamvc.tutorials.xmlc.Visibility1HTML) xmlcFactory.create(org.barracudamvc.tutorials.xmlc.Visibility1HTML.class);
        page1.getElementItem2().setAttribute("visdom","false");
        page1.getElementItem4().setAttribute("visdom","false");
        sw = new StringWriter(5000);
        try {
            dw.write(page1, sw);
        } catch (IOException e){
            System.out.println("Error writing DOM:"+e);
        }
        System.out.println(sw.toString());

        //load & print DOM page
System.out.println ("");
System.out.println ("Here 1");
        page2 = (org.barracudamvc.tutorials.xmlc.Visibility2HTML) xmlcFactory.create(org.barracudamvc.tutorials.xmlc.Visibility2HTML.class);
        page2.getElementItem2().setAttribute("visdom","false");
        page2.getElementItem4().setAttribute("visdom","false");
        sw = new StringWriter(5000);
        try {
            dw.write(page2, sw);
        } catch (IOException e){
            System.out.println("Error writing DOM:"+e);
        }
        System.out.println(sw.toString());
    }

*/

}
