/*
 * Copyright (C) 2001  Jacob Kjome [hoju@visi.com]
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: JivanDOMFactory.java 114 2005-12-09 15:51:51Z christianc $
 */
package org.barracudamvc.core.util.dom;

import java.io.IOException;
import java.io.InputStream;

import javax.servlet.ServletContext;

import org.apache.log4j.Logger;
import org.jivan.html.document.DocumentFactory;
import org.w3c.dom.Document;


/**
 * Jivan implementation of a DOMFactory. This class will load a DOM using
 * the Jivan DocumentFactory.
 */
public class JivanDOMFactory implements DOMFactory {

    /**
     * used for logging
     */
    protected static final Logger logger = Logger.getLogger(JivanDOMFactory.class.getName());

    /**
     * optional, used to load resources from oa servlet context
     * 
     * @see #setServletContext(ServletContext)
     */
    protected ServletContext servletContext;

    /**
     * arbitrary object used to synchronize upon
     */
    protected final Object sync = new Object();

    /**
     * used for flagging whether the jivanFactory has been initialized and
     * avoiding unnecessary synchronization.
     */
    protected boolean initialized = false;

    /**
     * Jivan DocumentFactory instance, stored so that it isn't re-created on
     * every request.
     */
    protected DocumentFactory jivanFactory;

    /**
     * This method is not supported by this dom factory and will
     * immediately throw an IOException if called!
     *
     * <p>Jivan does not use a wrapper class for documents like XMLC.
     * If you already have a Jivan DocumentManager, just use that directly.</p>
     */
    public Document getInstance(Class clazz) throws IOException {
        throw new IOException("Error: Unimplemented - Jivan DocumentFactory does not support creation of documents from a class.  If you already have a DocumentManager instance, use it directly");
    }

    /**
     * Load a document directly from file using Jivan.
     *
     * <p>The docPath can be provided in one of two forms: a path relative to
     * a servlet context such as "/WEB-INF/templates/foo.html", in which case
     * one would have to call {@link #setServletContext} before the first call
     * to this method, or a fully qualified URL (or URI) such as that returned
     * by {@link java.net.URL#toExternalForm}.</p>
     *
     * @see DOMFactory#getInstance(String)
     */
    public Document getInstance(String docPath) throws IOException {
        initFactory();
        if (servletContext != null && docPath.startsWith("/")) {
            InputStream in = servletContext.getResourceAsStream(docPath);
            if (in == null) throw new IOException("non-existent resource path: " + docPath);
            return jivanFactory.docManFor(docPath, in).getDocument();
        }
        return jivanFactory.docManFor(docPath).getDocument();
    }

    /**
     * Optional method to set the current servlet context.  Allows for
     * a docPath such as "/WEB-INF/templates/foo.html".  If the servlet context
     * is not set, a docPath of this type will fail resulting in getInstance()
     * returning an IOException.  Must be set before getInstance() is first
     * called in order to be used.
     *
     * @param iservletContext the current servlet context
     */
    public void setServletContext(ServletContext iservletContext) {
        this.servletContext = iservletContext;
    }

    private void initFactory() {
        if (initialized == false) {
            synchronized (sync) {
                logger.info("initializing a Jivan factory for returning HTML documents");
                jivanFactory = DocumentFactory.getInstance();
                initialized = true;
            }
        }
    }

}
