/*
 * Copyright (C) 2003  Christian Cryder [christianc@granitepeaks.com]
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: GenerateSSIs.java 114 2005-12-09 15:51:51Z christianc $
 */
package org.barracudamvc.taskdefs;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.FileSet;

/**
 * The purpose of this taskdef is to parse a source file (or fileset)
 * and create ssi's from it. This is useful for the mockup process, when
 * you want to edit your mockups in a single HTML file and then have
 * pieces of it automatically written out to .ssi files.
 *
 * <p>The tags are valid HTML comments that follow a format that looks
 * like this (where 'foo' can be the name of any ssi file):
 *
 * <p><!-- start foo.ssi -->
 * <br>...
 * <br><!-- end foo.ssi -->
 *
 * <p>When the taskdef sees these tags it will write the contents to the
 * target ssi file IFF the source file is newer than the ssi file or
 * force="true".
 *
 * <p>The taskdef also allows you to specify a touchpattern (defaults to *.shtml);
 * if the taskdef writes ssi files in that directory, it will also update the
 * timestamp on all the files in that directory that match the touch pattern. This
 * is useful for causing XMLC to automatically recompile these files the next time
 * you do a full build.
 *
 * @author <a href="mailto:christianc@granitepeaks.com">Christian Cryder</a>
 */
public class GenerateSSIs extends Task {
    protected File file = null;     // the source file
    protected List filesets = new ArrayList();

    protected String touchpattern = "**/*.shtml";
    protected String excludepattern = null;
    protected boolean force = false;
    protected int verbosity = Project.MSG_VERBOSE;

    /**
     * Sets a single source file to copy.
     */
    public void setFile(File file) {
        this.file = file;
    }

    /**
     * Specify a pattern of files to be touched if we end up rewriting
     * any .ssi files (this defaults to *.shtml")
     */
    public void setTouchpattern(String touchPattern) {
        this.touchpattern = touchPattern;
    }

    //csc_081805_1 - added
    /**
     * Specify a pattern of files to be excluded in the touch process 
     * (defaults to null)
     */
    public void setExcludepattern(String excludePattern) {
        this.excludepattern = excludePattern;
    }

    /**
     * Force the regeneration of SSIs
     */
    public void setForce(boolean force) {
        this.force = force;
    }

    /**
     * Used to force listing of all names of copied files.
     */
    public void setVerbose(boolean verbose) {
        if (verbose) {
            this.verbosity = Project.MSG_INFO;
        } else {
            this.verbosity = Project.MSG_VERBOSE;
        }
    }

    /**
     * Adds a set of files (nested fileset attribute).
     */
    public void addFileset(FileSet set) {
        filesets.add(set);
    }

    /**
     * Performs the copy operation.
     */
    public void execute() throws BuildException {
        // make sure we don't have an illegal set of options
        validateAttributes();

        List fileList = new ArrayList();

        // deal with the single file
        if (file!=null) {
            if (file.exists()) {
                fileList.add(file);
            } else {
                String message = "Could not find file "+file.getAbsolutePath()+" to generate ssi files from.";
                log(message);
                throw new BuildException(message);
            }
         // deal with the filesets
        } else  {
            for (int i=0; i<filesets.size(); i++) {
                FileSet fs = (FileSet) filesets.get(i);
                DirectoryScanner ds = fs.getDirectoryScanner(this.getProject());
//                File fromDir = fs.getDir(this.getProject());

                File baseDir = ds.getBasedir();
                String[] srcFiles = ds.getIncludedFiles();

                for (int j=0; j<srcFiles.length; j++) {
                    fileList.add(new File(baseDir, srcFiles[j]));
                }
            }
        }

        //now process our file list
        processFiles(fileList);
    }

    /**
     * Ensure we have a consistent and legal set of attributes, and set
     * any internal flags necessary based on different combinations
     * of attributes.
     */
    protected void validateAttributes() throws BuildException {
        if (file==null && filesets.size()==0) {
            throw new BuildException("Specify at least one source - a file or a fileset.");
        }

        if (file!=null && file.exists() && file.isDirectory()) {
            throw new BuildException("Use a fileset to generate ssis on directories.");
        }
    }

    /**
     * process the list of qualified files. We will read each of these files
     * looking for tags that indicate output should be redirected to a .ssi
     * file. The tags are valid HTML comments that follow a format that looks
     * like this:
     *
     * <p><!-- start foo.ssi -->
     * <br>...
     * <br><!-- end foo.ssi -->
     */
    protected void processFiles(List fileList) {
        Iterator it = fileList.iterator();
        int writeCnt = 0;
        int touchCnt = 0;
        Map touchedDirs = new HashMap();
        while (it.hasNext()) {
            File f = (File) it.next();
            boolean notified = false;
            log("Processing file: "+f, verbosity);

            try {
                BufferedReader in = new BufferedReader(new FileReader(f));
                BufferedWriter out = null;
                String nextLine = null;
                String targetSSI = null;
                while (true) {
                    nextLine=in.readLine();
                    if (nextLine==null) {
                        if (out!=null) {
                            out.flush();
                            out.close();
                            out = null;
                        }
                        break;
                    }

                    //if out==null then we are in between tags...check for
                    //the starting flag
                    if (out==null) {
                        //look for the start flag; if not present continue
                        int spos = nextLine.indexOf("<!-- start ");
                        if (spos<0) continue;
                        int epos = nextLine.indexOf(".ssi -->", spos);
                        if (epos<0) continue;

                        //get the target ssi name
                        targetSSI = nextLine.substring(spos+11, epos+4);
//                        log("Start tag for: "+targetSSI, verbosity);

                        //look for the target ssi file
                        File ssi = new File(f.getParent(), targetSSI);
                        if (force || !ssi.exists() || f.lastModified()>ssi.lastModified()) {
                            if (!notified) {
                                log("File "+f+" has been modified...");
                                notified = true;
                                touchedDirs.put(f.getParent(), f.getParentFile());
                            }
                            log("...Writing "+ssi, verbosity);
                            writeCnt++;
                            out = new BufferedWriter(new FileWriter(ssi));
//csc_012804_1 - don't include the ssi comment id tag in the generated file
//csc_012804_1                            out.write(nextLine.substring(spos));
                            out.write(nextLine.substring(epos+8));    //csc_012804_1
                            out.newLine();
                        }

                    //otherwise check for an ending flag and write the output
                    } else {
                        //look for the end tag; if not present continue
                        String eflg = "<!-- end "+targetSSI+" -->";
                        int spos = nextLine.indexOf(eflg);
                        if (spos<0) {
                            out.write(nextLine);
                            out.newLine();
                        } else {
//                            log("End tag for: "+targetSSI, verbosity);
//csc_012804_1 - don't include the ssi comment id tag in the generated file
//csc_012804_1                            out.write(nextLine.substring(0, spos+eflg.length()));
                            out.write(nextLine.substring(0, spos));   //csc_012804_1
                            out.newLine();
                            out.flush();
                            out.close();
                            out = null;
                            continue;
                        }
                    }
                }
                in.close();

            } catch (IOException e) {
                log("Unable to read file:"+f, Project.MSG_ERR);
                log("IOException "+e, Project.MSG_ERR);
            }
        }

        //now touch all the appropriate files
        if (touchpattern!=null && touchedDirs.size()>0) {
            it = touchedDirs.values().iterator();
            while (it.hasNext()) {
                File baseDir = (File) it.next();
                log("Touching "+baseDir+", tocuhpattern=\""+touchpattern+"\", excludepattern=\""+excludepattern+"\"");
                FileSet fs = new FileSet();
                fs.setDir(baseDir);
                fs.setIncludes(touchpattern);
                if (excludepattern!=null) fs.setExcludes(excludepattern);   //csc_081805_1
                DirectoryScanner ds = fs.getDirectoryScanner(this.getProject());
                String[] touchFiles = ds.getIncludedFiles();
                long millis = System.currentTimeMillis();
                for (int j=0; j<touchFiles.length; j++) {
                    File tf = new File(baseDir, touchFiles[j]);
                    log("...Touching "+tf, verbosity);
                    tf.setLastModified(millis);
                    touchCnt++;
                }
            }
        }

        //log it
        if (writeCnt>0 || touchCnt>0) {
            log("Created "+writeCnt+" .ssi files, touched "+touchCnt+" "+touchpattern+" files.");
        }
    }
}
